Require Import Coqlib.
Require Import Maps.
Require Import Values.
Require Import Integers.
Require Import ZArith.
Require Import VTDefs.
Require Import LemmaMaths.
Require Import Psatz.
Require Import LemmaVTMap.
Require Import LemmaInterference.
Require Import LemmaOtherWorkload.
Require Import LemmaOtherWorkload1.


Open Scope Z_scope.


Section SCHED_SELF.

  Lemma count_occurrence_self_is_budget_single_period_aux_precise:
    forall N ts nrange conf id T C te,
      0 <= id < N ->
      0 <= ts ->
      1 < T ->
      0 < C ->
      ts mod T = 0 ->
      Z.of_nat nrange <= T ->
      ts + T <= te ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      count_type_occurrence_aux N N ts nrange conf id te Sched_self = Z.min C (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) - ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
  Proof.
    intros * Hid Hts HT HC Hmod Hnrange Hte Hconf.
    induction nrange.
    {
      simpl.
      replace (ts + 0) with ts by omega.
      rewrite Z.min_r; omega.
    }
    {
      assert (H1: ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1). xomega.
      rewrite H1.
      simpl.
      assert (HN: N = Z.of_nat (Z.to_nat N)). rewrite Z2Nat.id; omega.
      assert (Hmono: monotonic_continuous (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf))).
      {
        apply TAC_pool_all_monotonic_continuous; auto; try xomega.
      }        
      assert (H2: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) \/
                  ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) + 1).
      {
        replace (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange)) with (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange)).
        - apply Hmono; omega.
        - rewrite vertical_map_canonical; omega.
      }

      assert (Hrange': Z.of_nat nrange <= T). xomega.
      specialize (IHnrange Hrange'). clear Hrange'.

      assert (H4: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                           (ts + Z.of_nat nrange) <=
             ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf)
                      (ts + Z.of_nat nrange + 1) <=
             ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                      (ts + Z.of_nat nrange) + 1).
      {
        destruct H2; omega.
      }
      
      destruct (zeq
         (highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) N) as [Hidle|Hreal].
      {
        (* idle, task must have been finished *)
        destruct (vt_map_not_finished (ts + Z.of_nat nrange) T C (ZMap.get id ((TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)))) eqn:Hfinish.
        {
          exfalso.
          assert (Hreal: 0 <= highest_not_finished_map N (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) < N).
          {
            eapply exists_ready_task_highest_not_finished_map_schedule_something; eauto.
          }
          omega.
        }
        {
          unfold vt_map_not_finished in Hfinish.
          rewrite divide_plus_equal in Hfinish; try omega.
          rewrite mod_0_divide_mul in Hfinish; try omega.

          replace (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) ts) with (ZMap.get id (TAC_pool_at_t (ts) (Z.to_nat N) conf) ts) in Hfinish.
          - destruct (zlt
              (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                 (ts + Z.of_nat nrange) -
               ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts)
              C); try discriminate.
            rewrite Z.min_l in IHnrange; try omega.
            rewrite Z.min_l; try omega.            
          - symmetry.
            rewrite vertical_map_canonical; omega.
        }
      }
      {
        (* not idle *)
        destruct (ZMap.get
                    (highest_not_finished_map N (ts + Z.of_nat nrange)
                                              (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                              (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) conf) as [T0 C0] eqn:Hconf0.
        destruct (sched_type_eq_dec
         (get_sched_type
            (highest_not_finished_map N (ts + Z.of_nat nrange)
               (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) T0 id
            (ts + Z.of_nat nrange) te) Sched_self) as [Hself | Hothers]; try omega.
        {
          (* self is scheduled *)
          pose proof (get_sched_type_is_self _ _ _ _ _ Hself) as Hideq.
          
          assert (H5: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) = ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) + 1).
          {
            apply vt_increase_for_the_scheduled_task; auto; try omega.
            rewrite <- HN.
            auto.
          }
          assert (H6: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) - ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts < C).
          {
            assert (H7: vt_map_not_finished (ts + Z.of_nat nrange) T C (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)) = true).
            {
              apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf:= conf); auto; try omega.
              rewrite <- HN.
              auto.
            }
            unfold vt_map_not_finished in H7.
            rewrite divide_plus_equal in H7; try omega.
            rewrite mod_0_divide_mul in H7; try omega.
            assert (H8: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) ts = ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
            { rewrite vertical_map_canonical; omega. }
            rewrite H8 in H7.
            destruct (zlt
                        (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange) -
                         ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts) C); try omega; try discriminate.
          }
          rewrite H5.
          rewrite Z.min_r in IHnrange; try omega.
          rewrite Z.min_r; try omega.      
        }
        {
          (* others are scheduled *)
          destruct (vt_map_not_finished (ts + Z.of_nat nrange) T C (ZMap.get id ((TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)))) eqn:Hfinish.
          {
            (* preempted *)
            assert (Hvt: ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange + 1) (Z.to_nat N) conf) (ts + Z.of_nat nrange + 1) =
                    ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) (ts + Z.of_nat nrange)).
            {
              apply vt_stagnates_if_task_not_scheduled with (T:=T) (C:=C); auto; try omega.
              - rewrite Z2Nat.id; try omega.
                assert (0 <=
   highest_not_finished_map N (ts + Z.of_nat nrange)
     (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
     (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf) <= N).
                {
                  apply highest_not_finished_map_res_bound; auto.
                  + omega.
                  + remember (Z.to_nat N) as Nnat.
                    replace N with (Z.of_nat Nnat).
                    apply real_priqueue_sound.
                }
                omega.
              - rewrite Z2Nat.id; try omega.
                unfold get_sched_type in Hothers.
                destruct (zeq
                  (highest_not_finished_map N (ts + Z.of_nat nrange)
                     (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                     (real_priqueue (Z.to_nat N) (ts + Z.of_nat nrange) conf)) id); try congruence.
            }

            rewrite Hvt.
            auto.
          }
          {
            (* already finished *)
            unfold vt_map_not_finished in Hfinish.
            rewrite divide_plus_equal in Hfinish; try omega.
            rewrite mod_0_divide_mul in Hfinish; try omega.

            replace (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf) ts) with (ZMap.get id (TAC_pool_at_t (ts) (Z.to_nat N) conf) ts) in Hfinish.
            - destruct (zlt
                          (ZMap.get id (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N) conf)
                                    (ts + Z.of_nat nrange) -
                           ZMap.get id (TAC_pool_at_t ts (Z.to_nat N) conf) ts)
                          C); try discriminate.
              rewrite Z.min_l in IHnrange; try omega.
              rewrite Z.min_l; try omega.            
            - symmetry.
              rewrite vertical_map_canonical; omega.
          }
        }
      }
    }          
  Qed.
  
  Lemma count_occurrence_self_bounded_equals_budget_single_period:
    forall N ts conf id T C te,
      0 <= id < N ->
      0 <= ts ->      
      1 < T ->
      0 < C ->
      ts mod T = 0 ->
      ts + T <= te ->
      system_schedulable N conf ->      
      ZMap.get id conf = mkPrioConfigValid T C ->
      count_type_occurrence N N ts (ts + T) conf id te Sched_self = C.
  Proof.
    intros.
    unfold count_type_occurrence.    
    assert (Z.of_nat (Z.to_nat (ts + T - ts)) <= te).
    { rewrite Z2Nat.id; omega. }
    replace (ts + T - ts) with T by omega.
    replace te with (ts + te - ts) by omega.

    rewrite count_occurrence_self_is_budget_single_period_aux_precise with (T:=T) (C:=C); auto; try omega.
    - rewrite Z2Nat.id; try omega.
      rewrite Z.min_l; auto.
      apply system_schedulable_implies_period_schedulable; auto.
    - rewrite Z2Nat.id; omega.
  Qed.

  Lemma count_occurrence_self_more_than_budget_multiple_periods_aux:
    forall N ts conf id T C te k,
      0 <= k ->
      0 <= id < N ->
      0 <= ts ->      
      1 < T ->
      0 < C ->
      te mod T = 0 ->
      ts + k*T <= te ->
      system_schedulable N conf ->      
      ZMap.get id conf = mkPrioConfigValid T C ->
      k* C <= count_type_occurrence N N ts te conf id te Sched_self.
  Proof.
    intros * Hk Hid Hts HT HC Hmod Hte Hschedulable Hconf.    
    remember (Z.to_nat k) as nk.
    assert (Hkval: k = Z.of_nat nk).
    {
      rewrite Heqnk.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hkval in *. clear Hkval Heqnk.
    revert Hts Hte.
    revert ts.
    
    induction nk.
    {
      intros.
      simpl.
      apply count_type_occurrence_nonneg.
    }
    {
      intros * Hts Hte.
      rewrite <- count_type_occurrence_composition with (t2 := te - Z.of_nat nk * T); try xomega; try nia.
      assert (Hv1: Z.of_nat nk * C <= count_type_occurrence N N (te - Z.of_nat nk * T) te conf id te Sched_self).
      {
        apply IHnk; auto; try omega.
        nia.
      }

      assert (Hv2: C <= count_type_occurrence N N ts (te - Z.of_nat nk * T) conf id te Sched_self).
      {
        rewrite <- count_type_occurrence_composition with (t2 := te - Z.of_nat nk * T - T); try xomega; try nia.
        assert (count_type_occurrence N N (te - Z.of_nat nk * T - T) 
                                      (te - Z.of_nat nk * T) conf id te Sched_self = C).
        {
          replace (te - Z.of_nat nk * T) with (te - Z.of_nat nk * T - T + T) at 2 by omega.
          apply count_occurrence_self_bounded_equals_budget_single_period; auto; try nia.
          replace (te - Z.of_nat nk * T - T) with (te - (Z.of_nat nk + 1) * T) by lia.
          rewrite Zminus_mod.
          rewrite Hmod.
          rewrite Z_mod_mult.
          auto.
        }

        assert (0 <= count_type_occurrence N N ts (te - Z.of_nat nk * T - T) conf id te Sched_self).
        {
          apply count_type_occurrence_nonneg; auto.
        }
        nia.
      }

      nia.
    }    
  Qed.

  Lemma count_occurrence_self_more_than_budget_multiple_periods:
    forall N ts conf id T C te,
      0 <= id < N ->
      0 <= ts <= te ->      
      1 < T ->
      0 < C ->
      te mod T = 0 ->
      system_schedulable N conf ->      
      ZMap.get id conf = mkPrioConfigValid T C ->
      (te - ts) / T * C <= count_type_occurrence N N ts te conf id te Sched_self.
  Proof.
    intros.

    assert (H10: 0 <= (te - ts) / T).
    {
      apply Z_div_pos; omega.
    }
    assert (H11: (te - ts) / T * T <= te - ts).
    {
      apply divide_mul_le; omega.
    }
    
    apply count_occurrence_self_more_than_budget_multiple_periods_aux with (T:=T); auto; try omega.
  Qed.  

End SCHED_SELF.


Section SCHED_INTERFERENCE.

  Lemma count_single_task_interference_equals_budget_in_one_period:
    forall N ts conf toid i Ti Ci todeadline,
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->
      i <> toid ->
      ts mod Ti = 0 ->
      priEntry_lt (PriEntry (ts + Ti) i) (PriEntry todeadline toid) = true ->
      system_schedulable N conf ->
      count_single_task_type_occurrence N N i ts (ts + Ti) conf toid todeadline Sched_interference = Ci.
  Proof.
    intros * Hts Hconf HTf HCf Hid Hidbound Hmod Hpri Hschedulable.
    remember (Z.to_nat Ti) as nT.
    assert (H1: Ti = Z.of_nat nT).
    { rewrite HeqnT. rewrite Z2Nat.id; omega. }
    rewrite H1.
    
    rewrite count_single_task_interference_value_in_one_period_aux with (Ti:=Ti) (Ci:=Ci); auto; try omega.
    rewrite Z.min_l; auto.
    - rewrite <- H1.
      apply system_schedulable_implies_period_schedulable; auto; try omega.
  Qed.

  Lemma count_single_task_interference_equals_workload_subtotal_aux1_aux:
    forall N ts conf toid n i Ti Ci todeadline,
      ts < ts + Z.of_nat n * Ti <= todeadline ->
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->      
      ts mod Ti = 0 ->
      i <> toid ->
      priEntry_lt (PriEntry (ts + Ti * Z.of_nat n) i) (PriEntry todeadline toid) = true ->
      system_schedulable N conf ->
      count_single_task_type_occurrence N N i ts (ts + Ti * Z.of_nat n) conf toid todeadline Sched_interference = Ci * Z.of_nat n.
  Proof.
    intros * Htod Hts Hconf HT HC Hi Hmod Hto Hpri Hschedulable.
    induction n.
    {
      simpl in Htod.
      omega.
    }
    {
      assert (H1: Z.of_nat (S n) = Z.of_nat n + 1). xomega.
      assert (H2: ts + Ti * Z.of_nat (S n) = ts + Ti * Z.of_nat n + Ti). lia.
      rewrite H2 in *. clear H2.

      destruct (zeq (Z.of_nat n) 0) as [Heq | Hneq].
      {
        rewrite H1 in *.
        rewrite Heq in *.
        replace (ts + Ti * 0) with ts in * by omega.
        replace (Ci * (0 + 1)) with Ci by nia.
        apply count_single_task_interference_equals_budget_in_one_period; auto.
      }
      {                
        assert (H3: ts <= ts + Ti * Z.of_nat n <= ts + Ti * Z.of_nat n + Ti). nia.
        pose proof (count_single_task_type_occurrence_composition N N i _ _ _ conf toid todeadline Sched_interference H3) as H4.
        rewrite <- H4. clear H4.

        replace (Ci * Z.of_nat (S n)) with (Ci * Z.of_nat n + Ci) by nia.

        assert (H5: ts < ts + Z.of_nat n * Ti <= todeadline). nia.
        assert (H6: priEntry_lt (PriEntry (ts + Ti * Z.of_nat n) i) (PriEntry todeadline toid) = true).
        {
          unfold priEntry_lt.
          rewrite zlt_true; auto.
          rewrite H1 in Htod.         
          nia.
        }
        specialize (IHn H5 H6).
        assert (H: forall a b c d, a = c -> b = d -> a + b = c + d). intros. omega.
        apply H; auto.

        apply count_single_task_interference_equals_budget_in_one_period; auto; try omega.
        - rewrite Z.mul_comm.
          rewrite Z_mod_plus_full.
          auto.
      }
    }
  Qed.

  Lemma count_single_task_interference_equals_workload_subtotal_aux1:
    forall N ts conf toid n i Ti Ci todeadline,
      ts + Z.of_nat n * Ti <= todeadline ->
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->      
      ts mod Ti = 0 ->
      i <> toid ->
      priEntry_lt (PriEntry (ts + Ti * Z.of_nat n) i) (PriEntry todeadline toid) = true ->
      system_schedulable N conf ->
      count_single_task_type_occurrence N N i ts (ts + Ti * Z.of_nat n) conf toid todeadline Sched_interference = Ci * Z.of_nat n.
  Proof.
    intros.
    destruct (zeq (Z.of_nat n) 0) as [Heq | Hneq].
    {
      rewrite Heq.
      replace (ts + Ti * 0) with ts by omega.
      rewrite count_single_task_type_occurance_0_when_empty_duration.
      omega.
    }
    {
      apply count_single_task_interference_equals_workload_subtotal_aux1_aux; auto.
      nia.
    }
  Qed.  

  Lemma count_single_task_interference_equals_workload_subtotal_start_aligned:
    forall N ts conf toid T i Ti Ci,
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti <= T ->
      0 < Ci ->
      0 <= i < N ->      
      i <> toid ->
      ts mod Ti = 0 ->
      system_schedulable N conf ->
      count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts + T) Sched_interference =
      workload_whole toid i Ti Ci ts (ts + T).
  Proof.
    intros * Hts Hconf HT HC Hibound Hi Hmod Hschedulable.
    unfold workload_whole.
    rewrite zlt_false; try omega.
    rewrite Hmod.
    rewrite zeq_true.
    destruct (zeq ((ts + T) mod Ti) 0) as [Heq | Hneq].
    {
      assert (H1: 0 < T / Ti).
      {
        apply div_pos; omega.
      }
      remember (Z.to_nat (T / Ti)) as n.
      assert (H2: T / Ti = Z.of_nat n).
      {
        rewrite Heqn. rewrite Z2Nat.id; omega.
      }
      assert (H3: T = Ti * Z.of_nat n).
      {
        rewrite <- H2.
        rewrite Z.mul_comm.
        rewrite mod_0_divide_mul; try omega.
        rewrite <- Z.add_mod_idemp_l in Heq; try omega.
        rewrite Hmod in Heq.
        replace (0 + T) with T in Heq by omega.
        auto.
      }
      
      destruct (zlt i toid) as [Hlt|Hge].
      {
        rewrite mod_0_divide_mul; try omega.
        replace (ts + T - ts) with T by omega.
        rewrite H2. rewrite Z.mul_comm.
        rewrite H3.
        apply count_single_task_interference_equals_workload_subtotal_aux1; auto; try omega.
        - nia.
        - unfold priEntry_lt.
          rewrite zlt_false; try omega.
          rewrite zeq_true.
          rewrite zlt_true; auto.          
      }
      {
        replace (ts + T - Ti - ts) with ((Z.of_nat n - 1) * Ti) by nia.
        rewrite Z_div_mult_full; try omega.
        assert (H10: count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts + T) Sched_interference = count_single_task_type_occurrence N N i ts (ts + T - Ti) conf toid (ts + T) Sched_interference + count_single_task_type_occurrence N N i (ts + T - Ti) (ts + T) conf toid (ts + T) Sched_interference).
        {
          symmetry.
          apply count_single_task_type_occurrence_composition; try omega.
        }
        rewrite H10.
        remember (pred n) as n'.
        assert (H11: n = S n'). xomega.
        assert (H12: Z.of_nat n' = Z.of_nat n - 1). omega.
        assert (H14: T - Ti = Z.of_nat n' * Ti). nia.
        replace (ts + T - Ti) with (ts + Ti * Z.of_nat n') by nia.
        rewrite <- H12.

        assert (H: forall a b c, a = c -> b = 0 -> a + b = c). intros. omega.
        apply H.
        - replace (Z.of_nat n' * Ci) with (Ci * Z.of_nat n') by lia.
          apply count_single_task_interference_equals_workload_subtotal_aux1; auto; try omega.
          + unfold priEntry_lt.
            rewrite zlt_true; auto.
            nia.
        - apply count_single_task_interference_0_aux2 with (Ti:=Ti) (Ci:=Ci); auto; try nia.
          + rewrite Zplus_mod.
            rewrite Hmod.
            rewrite Z.mul_comm.
            rewrite Z_mod_mult.
            simpl.
            rewrite Z.mod_0_l; omega.
          + unfold priEntry_lt.
            assert (H15: ts + Ti * Z.of_nat n' + Ti = ts + T). nia.
            rewrite H15.
            rewrite zlt_false; try omega.
            rewrite zeq_true.
            rewrite zlt_false; auto.            
      }
    }
    {
      assert (H1: ts <= (ts + T) / Ti * Ti).
      {
        apply mod_not_0_adjust_to_smaller_period; omega.
      }
      assert (H2: (ts + T) / Ti * Ti <= ts + T).
      {
        apply divide_mul_le; omega.
      }
      assert (H5: ((ts + T) / Ti * Ti) mod Ti = 0).
      {
        rewrite Z_mod_mult. auto.
      }
      assert (H6: ts + T < (ts + T) / Ti * Ti + Ti).
      { apply x_lt_floor_plus; omega. }
      
      assert (H3: count_single_task_type_occurrence N N i ts (ts + T) conf toid (ts + T) Sched_interference = count_single_task_type_occurrence N N i ts ((ts + T) / Ti * Ti) conf toid (ts + T) Sched_interference + count_single_task_type_occurrence N N i ((ts + T) / Ti * Ti) (ts + T) conf toid (ts + T) Sched_interference).
      {
        symmetry.
        apply count_single_task_type_occurrence_composition; try omega.
      }
      rewrite H3. clear H3.

      assert (H4: count_single_task_type_occurrence N N i ((ts + T) / Ti * Ti) 
                                                    (ts + T) conf toid (ts + T) Sched_interference = 0).
      {
        apply count_single_task_interference_0_aux2 with (Ti:=Ti) (Ci:=Ci); auto; try nia.
        - unfold priEntry_lt.
          rewrite zlt_false; try omega.
          rewrite zeq_false; try omega.
          auto.
      }
      rewrite H4.
      rewrite Z.add_0_r.

      assert (H7: 0 <= ((ts + T) / Ti * Ti - ts) / Ti).
      {
        apply Z_div_pos; omega.
      }
      
      remember (Z.to_nat (((ts + T) / Ti * Ti - ts) / Ti)) as n.
      assert (H8: ((ts + T) / Ti * Ti - ts) / Ti = Z.of_nat n).
      { rewrite Heqn. rewrite Z2Nat.id; omega. }
      rewrite H8.
      replace (Z.of_nat n * Ci) with (Ci * Z.of_nat n) by nia.

      assert (H9: (ts + T) / Ti * Ti = ts + Ti * Z.of_nat n).
      {
        replace (Ti * Z.of_nat n) with (Z.of_nat n * Ti) by nia.
        rewrite <- H8.
        symmetry.
        rewrite mod_0_divide_mul; try omega.
        rewrite Zminus_mod.
        rewrite Hmod, H5.
        simpl.
        rewrite Z.mod_0_l; omega.
      }
      rewrite H9.

      assert (H10: ts + T - (ts + T) / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }
      
      assert (H11: ts + Z.of_nat n * Ti < ts + T). nia.
        
      apply count_single_task_interference_equals_workload_subtotal_aux1; auto; try omega.
      - unfold priEntry_lt.
        rewrite zlt_true; try omega.
        auto.
    }
  Qed.

  Lemma count_single_task_interference_equals_workload_subtotal_start_aligned_generalized:
    forall N ts te conf toid i Ti Ci,
      0 <= ts <= te ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      0 <= i < N ->      
      i <> toid ->
      ts mod Ti = 0 ->
      system_schedulable N conf ->
      count_single_task_type_occurrence N N i ts te conf toid te Sched_interference =
      workload_whole toid i Ti Ci ts te.
  Proof.
    intros * Hts Hconf HT HC Hibound Hi Hmod Hschedulable.
    
    destruct (zlt (te - ts) Ti) as [Htesmall | Htelarge].
    {
      unfold workload_whole.
      rewrite zlt_true; auto.
      rewrite count_single_task_interference_0_aux2 with (Ti:=Ti) (Ci:=Ci); auto; try omega.
      - unfold priEntry_lt.
        rewrite zlt_false; try omega.
        rewrite zeq_false; try omega.
        auto.
    }
    {
      assert (H1: te = ts + (te - ts)). omega.
      rewrite H1.
      apply count_single_task_interference_equals_workload_subtotal_start_aligned; auto; try omega.
    }
  Qed.

  Lemma count_single_task_interference_equals_remain_aux3:
    forall N i Ti Ci ts toid todeadline conf,
      0 <= ts ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      0 < Ti ->
      0 < Ci ->
      i <> toid ->
      0 <= i < N ->
      ts mod Ti <> 0 ->
      priEntry_lt (PriEntry (ts / Ti * Ti + Ti) i) (PriEntry todeadline toid) = true ->
      system_schedulable N conf ->
      count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference =
      workload_remain toid i Ti Ci ts todeadline (ZMap.get i (TAC_pool_at_t todeadline (Z.to_nat N) conf)).
  Proof.
    intros * Hts Hconf HT HC Hi Hibound Hmod Hpri Hschedulable.
    assert (H1: ts / Ti * Ti < ts).
    {
      assert (ts - ts / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }
      omega.
    }
    
    assert (H2: ts < ts / Ti * Ti + Ti).
    {
      apply x_lt_floor_plus; omega.
    }
      
    assert (H3: 0 <= ts / Ti * Ti).
    {
      apply Z.mul_nonneg_nonneg; try omega.
      apply Z_div_pos; omega.
    }
      
    assert (H4: (ts / Ti * Ti) mod Ti = 0).
    {
      apply Z_mod_mult.
    }
          
    assert (H: count_single_task_type_occurrence N N i (ts / Ti * Ti) ts conf toid todeadline Sched_interference + count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference = count_single_task_type_occurrence N N i (ts / Ti * Ti) (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference).
    {
        apply count_single_task_type_occurrence_composition; try omega.
    }

    replace (count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference) with (count_single_task_type_occurrence N N i (ts / Ti * Ti) 
                                                                                                                                                             (ts / Ti * Ti + Ti) conf toid todeadline Sched_interference - count_single_task_type_occurrence N N i (ts / Ti * Ti) ts conf toid todeadline Sched_interference) by omega.
    clear H.

    remember (Z.to_nat Ti) as nT.
    assert (H10: ts / Ti * Ti + Ti = ts / Ti * Ti + Z.of_nat nT).
    {
      rewrite HeqnT.
      rewrite Z2Nat.id; omega.
    }
    rewrite H10.
    assert (Hc1: count_single_task_type_occurrence N N i (ts / Ti * Ti) 
                                                   (ts / Ti * Ti + Z.of_nat nT) conf toid todeadline Sched_interference = Z.min Ci (ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Z.of_nat nT) (Z.to_nat N) conf) (ts / Ti * Ti + Z.of_nat nT) - ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti))).
    {
      apply count_single_task_interference_value_in_one_period_aux with (Ti:=Ti); auto; try omega.
    }
    rewrite Hc1. clear Hc1.

    remember (Z.to_nat (ts - ts / Ti * Ti)) as nT2.
    assert (H11: ts = ts / Ti * Ti + Z.of_nat nT2).
    {
      rewrite HeqnT2.
      rewrite Z2Nat.id; omega.
    }
    assert (Hc2: count_single_task_type_occurrence N N i (ts / Ti * Ti) ts conf toid todeadline Sched_interference = Z.min Ci (ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Z.of_nat nT2) (Z.to_nat N) conf) (ts / Ti * Ti + Z.of_nat nT2) - ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti))).
    {
      remember (ts / Ti * Ti) as k.
      rewrite H11.
      rewrite Heqk in *. clear Heqk.
      apply count_single_task_interference_value_in_one_period_aux with (Ti:=Ti); auto; try omega.
    }
    rewrite Hc2. clear Hc2.

    rewrite <- H10.
    rewrite <- H11.
    
    unfold workload_remain.
    rewrite zeq_false; auto.
    rewrite Hpri.

    assert (Hdeadline: ts / Ti * Ti + Ti <= todeadline).
    {
      eapply priEntry_lt_deadline_le; eauto.
    }
    
    assert (Hc3: ZMap.get i (TAC_pool_at_t todeadline (Z.to_nat N) conf) ts = ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts).
    {
      rewrite vertical_map_canonical; omega.
    }
    rewrite Hc3. clear Hc3.
    
    assert (Hc4: ZMap.get i (TAC_pool_at_t todeadline (Z.to_nat N) conf) (ts / Ti * Ti) = ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)).
    {
      rewrite vertical_map_canonical; omega.
    }
    rewrite Hc4. clear Hc4.

    assert (Hvt1: ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts <= ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Ti) (Z.to_nat N) conf) (ts / Ti * Ti + Ti)).
    {
      apply TAC_pool_at_t_monotonic_two_points; try omega.
    }

    assert (Hv3: Ci <= ZMap.get i (TAC_pool_at_t (ts / Ti * Ti + Ti) (Z.to_nat N) conf) (ts / Ti * Ti + Ti) -
                      ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)).
    {
      apply system_schedulable_implies_period_schedulable; auto; try omega.
    }
    rewrite Z.min_l; try omega.
    
    destruct (zle (ZMap.get i (TAC_pool_at_t ts (Z.to_nat N) conf) ts -
                   ZMap.get i (TAC_pool_at_t (ts / Ti * Ti) (Z.to_nat N) conf) (ts / Ti * Ti)) Ci).
    {
      rewrite Z.max_r; try omega.
      rewrite Z.min_r; try omega.
    }
    {
      rewrite Z.min_l; try omega.
      rewrite Z.max_l; try omega.
    }
  Qed.

  Lemma small_t_interference_not_possible:
    forall ts te i Ti toid N conf t,
      0 <= ts ->
      1 < Ti ->
      t <= ts / Ti * Ti + Ti ->
      priEntry_lt (PriEntry (ts / Ti * Ti + Ti) i) (PriEntry te toid) = false ->
      single_task_ty_not_possible_in_range ts t N conf i Ti toid te Sched_interference.
  Proof.
    intros * Hts HTi Hpri Ht1.
    unfold single_task_ty_not_possible_in_range.
    intros * Ht0 Hsched.
    unfold get_sched_type.
    destruct (zeq i toid); try congruence.

    assert (H1: ts / Ti * Ti <= ts).
    {
      apply divide_mul_le; omega.
    }
    assert (H2: 0 <= ts / Ti * Ti).
    {
      apply div_mul_nonneg; omega.
    }
    assert (H3: (ts / Ti * Ti) mod Ti = 0).
    {
      apply Z_mod_mult.
    }
    
    assert (Ht00: t0 / Ti * Ti = ts / Ti * Ti).
    {
      apply mod_lower_bound; try omega.
    }
    rewrite Ht00.
    rewrite Ht1.
    destruct (zeq (ts / Ti * Ti + Ti) te); try congruence.
  Qed.
    
  Lemma count_single_task_interference_equals_workload_subtotal:
    forall N ts te conf toid i Ti Ci,
      0 < N ->
      0 <= ts <= te ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      1 < Ti ->
      0 < Ci ->
      i <> toid ->
      0 <= i < N ->
      system_schedulable N conf ->
      count_single_task_type_occurrence N N i ts te conf toid te Sched_interference =
      workload_subtotal toid i Ti Ci ts te
                        (ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)).
  Proof.
    intros * HN Hts Hconf HTf HCf Hid Hibound Hschedulable.
    destruct (zeq (ts mod Ti) 0) as [Hmod|Hmod].
    {
      unfold workload_subtotal.
      rewrite workload_remain_0_start; auto.

      rewrite Z.add_0_l.
      replace te with (ts + (te - ts)) by omega.
      apply count_single_task_interference_equals_workload_subtotal_start_aligned_generalized; auto; try omega.
    }
    {
      unfold workload_subtotal.
      assert (H1: ts < ts / Ti * Ti + Ti).
      {
        apply x_lt_floor_plus; omega.
      }
      assert (H2: ts - ts / Ti * Ti >= 1).
      {
        apply z_gt_mod; omega.
      }

      destruct (zle (ts / Ti * Ti + Ti) te).
      {
        assert (H3: count_single_task_type_occurrence N N i ts te conf toid te Sched_interference = count_single_task_type_occurrence N N i ts (ts / Ti * Ti + Ti) conf toid te Sched_interference + count_single_task_type_occurrence N N i (ts / Ti * Ti + Ti) te conf toid te Sched_interference).
        {
          symmetry.
          apply count_single_task_type_occurrence_composition; try omega.
        }
        rewrite H3. clear H3.

        assert (H4: count_single_task_type_occurrence N N i (ts / Ti * Ti + Ti) 
                                                      te conf toid te Sched_interference = workload_whole toid i Ti Ci (ts / Ti * Ti + Ti) te).
        {
          apply count_single_task_interference_equals_workload_subtotal_start_aligned_generalized; auto; try omega.
          - rewrite Z.add_comm.
            rewrite Z_mod_plus_full.
            apply Z_mod_same_full.
        }

        rewrite workload_whole_left_align; try omega.

        assert (forall a b c d, b = d -> a = c -> a + b = c + d). intros. omega.
        apply H; auto.

        destruct (priEntry_lt (PriEntry (ts / Ti * Ti + Ti) i) (PriEntry te toid)) eqn:Hpri.
        {          
          apply count_single_task_interference_equals_remain_aux3; auto; omega.
        }
        {
          assert (Hw: workload_remain toid i Ti Ci ts te (ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)) = 0).
          {
            unfold workload_remain.            
            rewrite Hpri.
            destruct (zeq (ts mod Ti) 0); auto.
          }
          rewrite Hw.

          apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Ti) (C:=Ci); auto; try omega.
          - apply small_t_interference_not_possible; auto; try omega.
        }
      }
      {
        assert (Hpri: priEntry_lt (PriEntry (ts / Ti * Ti + Ti) i) (PriEntry te toid) = false).
        {
          unfold priEntry_lt.
          rewrite zlt_false; try omega.
          rewrite zeq_false; try omega.
          auto.
        }
        
        assert (Hw1: workload_remain toid i Ti Ci ts te (ZMap.get i (TAC_pool_at_t te (Z.to_nat N) conf)) = 0).
        {
          unfold workload_remain.
          destruct (zeq (ts mod Ti) 0); auto.
          rewrite Hpri.
          auto.
        }

        assert (Hw2: workload_whole toid i Ti Ci ts te = 0).
        {          
          unfold workload_whole.
          rewrite zlt_true; try omega.
        }

        rewrite Hw1, Hw2.
        simpl.

        apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Ti) (C:=Ci); auto; try omega.
        - apply small_t_interference_not_possible; auto; try omega.
      }
    }    
  Qed.

  Lemma count_occurrence_inteference_equals_aux:
    forall Nnat N0 id ts te conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 -> 
      0 <= ts <= te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      system_schedulable N0 conf ->
      count_type_occurrence2_aux Nnat N0 N0
                           ts te conf id te Sched_interference =
      workload_total2 Nnat id ts te conf (TAC_pool_at_t te (Z.to_nat N0) conf).
  Proof.
    intros * HNnat HN0 Hts Hconfvalid Hschedulable.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid').
      simpl.

      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.        
        rewrite count_single_task_self_sched_others_is_0; auto; try omega.
        congruence.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.        

        assert (H4: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
        assert (H2: 1 < Tf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        assert (H5: 0 < Cf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }

        assert (H: forall a b c d, a = d -> b = c -> a + b = c + d). intros. omega.
        apply H; auto.

        apply count_single_task_interference_equals_workload_subtotal; auto.
        - omega.
      }
    }      
  Qed.
  
  Lemma count_occurrence_interference_equals:
    forall Nnat id ts te conf,
      0 < Z.of_nat Nnat ->
      0 <= ts <= te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      system_schedulable (Z.of_nat Nnat) conf ->
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                            ts te conf id te Sched_interference =
      workload_total2 Nnat id ts te conf
                      (TAC_pool_at_t te Nnat conf).
  Proof.
    intros * HN Hts Hconfvalid Hschedulable.
    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    assert (H2: TAC_pool_at_t te Nnat conf = TAC_pool_at_t te (Z.to_nat (Z.of_nat Nnat)) conf).
    { rewrite H1. auto. }
    rewrite H2.

    replace te with (ts + (te - ts)) by omega.
    apply count_occurrence_inteference_equals_aux; auto; try omega.     
  Qed.  

End SCHED_INTERFERENCE.



Section SCHED_LEFTOVER.
    
  Lemma count_occurrence_leftover_more_aux_v:
    forall Nnat N0 id ts te conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 -> 
      0 <= ts <= te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te) ->
      system_schedulable N0 conf ->
      count_type_occurrence2_aux Nnat N0 N0
                           ts te conf id te Sched_leftover >=
      workload_leftover_total Nnat id ts te conf.
  Proof.
    intros * HNnat HN0 Hts Hconfvalid Hlarge Hschedulable.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (Hlarge': forall i Tn Cn : Z,
                 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te).
      {
        intros.
        eapply Hlarge; eauto.
        omega.
      }      
      
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid' Hlarge').
      simpl.
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.
        rewrite count_single_task_self_sched_leftover_is_0; try omega.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.        

        assert (H4: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
        assert (H2: 1 < Tf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }
        assert (H5: 0 < Cf).
        {
          specialize (Hconfvalid _ _ _ H4 Hconff).
          omega.
        }

        destruct (is_leftover (Z.of_nat Nnat) Tf id ts te) eqn:Hleftover.
        {
          assert (H: forall a b c d, a >= d -> b >= c -> a + b >= c + d). intros. omega.
          apply H; auto.
          remember (Z.to_nat (te - ts)) as nT.
          assert (HH: te = ts + Z.of_nat nT).
          { rewrite HeqnT. rewrite Z2Nat.id; omega. }
          rewrite HH in *.

          assert (H6: te mod Tf = 0).
          {
            unfold is_leftover in Hleftover.
            rewrite <- HH in Hleftover.
            destruct (zle ts (te - Tf)); try congruence.
            destruct (zeq (te mod Tf) 0); try congruence.
          }
          
          rewrite count_single_task_self_sched_leftover_bounded_by_budget_aux_v with (Ti:=Tf) (Ci:=Cf); auto; try omega.
          - rewrite Z.min_l; auto; try omega.
            remember (ts + Z.of_nat nT - Tf) as t.
            replace (ts + Z.of_nat nT) with (t + Tf) by omega.

            assert (Ht: 0 <= t).
            {
              assert (H20: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
              specialize (Hlarge _ _ _ H20 Hconff).
              omega.
            }
            assert (Ht2: t mod Tf = 0).
            {
              rewrite Heqt.
              rewrite <- HH.
              rewrite Zminus_mod.
              rewrite H6.
              rewrite Z_mod_same_full.
              auto.
            }
            
            assert (Hv1: ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t (t + Tf) (Z.to_nat N0) conf) t =
                         ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t t (Z.to_nat N0) conf) t).
            {
              rewrite vertical_map_canonical; try omega.
            }
            rewrite Hv1.

            apply system_schedulable_implies_period_schedulable; auto; try xomega.
          - rewrite HeqnT. rewrite Z2Nat.id; auto; try omega.
            assert (H11: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
            specialize (Hlarge _ _ _ H11 Hconff).
            omega.
        }        
        {
          assert (H: 0 <= count_single_task_type_occurrence N0 N0 (Z.of_nat Nnat) ts 
                                                       te conf id te Sched_leftover).
          {
            apply count_single_task_type_occurrence_nonneg.
          }
          omega.
        }
      }
    }      
  Qed.
  
  Lemma count_occurrence_leftover_more:
    forall Nnat id ts te conf,
      0 < Z.of_nat Nnat ->
      0 <= ts <= te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn, 0 <= i < Z.of_nat Nnat ->
                  ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= te) ->
      system_schedulable (Z.of_nat Nnat) conf ->
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                           ts te conf id te Sched_leftover >=
      workload_leftover_total Nnat id ts te conf.
  Proof.
    intros * HN Hts Hconfvalid Hlarge Hschedulable.
    
    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    apply count_occurrence_leftover_more_aux_v with (ts:=ts); auto; try omega. 
  Qed.
  
End SCHED_LEFTOVER.



Section SCHED_OVERLAP.

  (* workload_remain also counts long overlaps, which are not accounted for
       in workload_overlapping_remain_total *)
  Lemma workload_overlapping_remain_total_le_ts:
    forall Nnat N0 id ts te te' conf,
      0 <= ts <= te ->
      te <= te' ->
      Z.of_nat Nnat <= N0 ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      workload_overlapping_total Nnat id ts te te' conf -
      workload_remain_total Nnat id te te' conf (TAC_pool_at_t te' (Z.to_nat N0) conf) <=
      workload_overlapping_remain_total Nnat id ts te te' te conf (TAC_pool_at_t te' (Z.to_nat N0) conf).
  Proof.
    intros * Hts Hte' HN Hconfvalid.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      simpl.
      
      assert (H3: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (H4: config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (HN': Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat HN' H4).
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.

      assert (H6: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
      assert (H7: 1 < Tn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }
      assert (H9: 0 < Cn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }

      destruct (is_overlapping (Z.of_nat Nnat) Tn id ts te te') eqn:Hoverlap.
      {
        rewrite is_overlapping_implies_remain with (d1:=ts); auto; try omega.

        replace (Cn + workload_overlapping_total Nnat id ts te te' conf -
   (Cn -
    Z.min Cn
      (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) te -
       ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tn * Tn)) +
    workload_remain_total Nnat id te te' conf (TAC_pool_at_t te' (Z.to_nat N0) conf))) with
            (workload_overlapping_total Nnat id ts te te' conf -
             workload_remain_total Nnat id te te' conf (TAC_pool_at_t te' (Z.to_nat N0) conf) +
            Z.min Cn
      (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) te - ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tn * Tn))) by omega.
        assert (H: forall a c b, a <= b -> a + c <= b + c). intros. omega.
        apply H; auto.
      }
      {
        assert (Hremain: 0 <= workload_remain id (Z.of_nat Nnat) Tn Cn te te' (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t te' (Z.to_nat N0) conf))).
        {
          apply workload_remain_nonneg.
        }
        omega.
      }
    }
  Qed.

  Lemma is_overlapping_then_single_task_type_count_le_ts:
    forall N0 fromid toid ts te te' Tf Cf conf nrange,
      0 < N0 ->
      0 <= ts <= te ->
      te <= te' ->
      te / Tf * Tf < ts + Z.of_nat nrange <= te ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf ->
      0 < Cf ->
      is_overlapping fromid Tf toid ts te te' = true ->
      fromid <> toid ->
      0 <= fromid < N0 ->
      Z.min Cf (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (ts + Z.of_nat nrange) - ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) ((ts + Z.of_nat nrange) / Tf * Tf)) <= count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid te Sched_overlap.
  Proof.
    intros * HN0 Hts Hte' Hrange Hconf HTf HCf Hoverlap Hid Hfrom.
    unfold count_single_task_type_occurrence.
    replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl in Hrange.
      unfold is_overlapping in Hoverlap.
      destruct (zle ts (te / Tf * Tf)); try omega; try congruence.
    }
    {
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      assert (H2: ts + Z.of_nat (S nrange) = (ts + Z.of_nat nrange) + 1). xomega.
      rewrite H2.
      assert (H3: (te / Tf * Tf) mod Tf = 0).
      { apply Z_mod_mult. }
      assert (H10: 0 <= te / Tf * Tf).
      {
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H11: 0 <= fromid < Z.of_nat (Z.to_nat N0)).
      { rewrite Z2Nat.id; omega. }
      assert (H12: (ts + Z.of_nat nrange + 1) mod Tf <> 0).
      {            
        apply a_between_floor_and_self_mod_not_0 with (c := te); omega.
      }
      assert (H15: 0 <= (ts + Z.of_nat nrange) / Tf * Tf).
      { 
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H20: (ts + Z.of_nat nrange) / Tf * Tf <= ts + Z.of_nat nrange).
      {
        apply divide_mul_le; omega.
      }
      assert (H21: te / Tf * Tf <= te).
      {
        apply divide_mul_le; omega.
      }
      
      simpl.

      destruct (zeq
         (highest_not_finished_map N0 (ts + Z.of_nat nrange)
            (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf) conf
            (real_priqueue (Z.to_nat N0) (ts + Z.of_nat nrange) conf)) fromid) as [Hsched|Hnot].
      {
        rewrite Hsched.
        rewrite Hconf.
        
        assert (Htype: get_sched_type fromid Tf toid (ts + Z.of_nat nrange) te = Sched_overlap).
        {
          apply is_overlapping_type_is_overlap_generalized_v with (ts:=ts) (te':=te'); auto; try omega.          
        }
        rewrite Htype.
        rewrite sched_type_eq_true.

        destruct (zeq (ts + Z.of_nat nrange) (te / Tf * Tf)) as [Heq | Hneq].
        {          
          rewrite Heq in *.
            
          assert (H0: count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid te Sched_overlap = 0).
          {
            apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Tf) (C:=Cf); auto.
            - omega.
            - rewrite Heq.
              apply is_overlapping_not_possible_with_small_t; auto; try omega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.
        
          assert (H4: (te / Tf * Tf + 1) / Tf * Tf = te / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H5: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tf * Tf + 1) = ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (te / Tf * Tf) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_increase_for_the_scheduled_task; auto.
            rewrite Z2Nat.id; omega.            
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          assert (H22: te / Tf * Tf < ts + Z.of_nat nrange <= te). omega.
          specialize (IHnrange H22).
          
          transitivity (Z.min Cf
               (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (ts + Z.of_nat nrange) -
                ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                         ((ts + Z.of_nat nrange) / Tf * Tf)) + 1); try omega.

          rewrite mod_equal; try omega.
          assert (H14: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                 (ts + Z.of_nat nrange + 1) =
                        ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                 (ts + Z.of_nat nrange) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.

            apply vt_increase_for_the_scheduled_task; auto.
            omega.
            rewrite Z2Nat.id; omega.            
          }

          rewrite H14.

          destruct (zlt (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (ts + Z.of_nat nrange) -
                         ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) ((ts + Z.of_nat nrange) / Tf * Tf)) Cf).
          {
            repeat rewrite Z.min_r; try omega.
          }
          {
            repeat rewrite Z.min_l; try omega.
          }
        }
      }
      {
        (* fromid not scheduled *)
        destruct (zeq (ts + Z.of_nat nrange) (te / Tf * Tf)) as [Heq | Hneq].
        {          
          rewrite Heq in *.
          
          assert (H0: count_single_task_type_occurrence N0 N0 fromid ts (ts + Z.of_nat nrange) conf toid te Sched_overlap = 0).
          {
            apply count_single_task_type_occurrence_ty_is_0_if_not_possible with (T:=Tf) (C:=Cf); auto.
            - omega.
            - rewrite Heq.
              apply is_overlapping_not_possible_with_small_t; auto; try omega.
          }
          unfold count_single_task_type_occurrence in H0.
          replace (ts + Z.of_nat nrange - ts) with (Z.of_nat nrange) in H0 by omega.
          rewrite Nat2Z.id in H0.
          rewrite H0.
          
          assert (H4: (te / Tf * Tf + 1) / Tf * Tf = te / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H6: vt_map_not_finished (te / Tf * Tf) Tf Cf
                                          (ZMap.get fromid (TAC_pool_at_t (te / Tf * Tf) (Z.to_nat N0) conf)) = true).
          {
            apply vt_map_not_finished_true_at_beginning; auto; omega.
          }
          
          assert (H5: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                               (te / Tf * Tf + 1) =
                      ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                               (te / Tf * Tf)).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
            - rewrite Z2Nat.id; try omega.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
            - rewrite Z2Nat.id; omega.
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          assert (H22: te / Tf * Tf < ts + Z.of_nat nrange <= te). omega.
          specialize (IHnrange H22).
          
          transitivity (Z.min Cf
               (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf) (ts + Z.of_nat nrange) -
                ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                         ((ts + Z.of_nat nrange) / Tf * Tf))); try omega.
          rewrite mod_equal; try omega.

          destruct (vt_map_not_finished (ts + Z.of_nat nrange) Tf Cf
     (ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf))) eqn:Hfinish.
          {
            assert (H5: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                 (ts + Z.of_nat nrange + 1) =
                   ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                            (ts + Z.of_nat nrange)).
            {
              rewrite vertical_map_canonical; try omega.
              symmetry.
              rewrite vertical_map_canonical; try omega.
              symmetry.
              apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
              - rewrite Z2Nat.id; try omega.
                apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
              - rewrite Z2Nat.id; omega.
            }
            rewrite H5.
            omega.
          }
          {
            unfold vt_map_not_finished in Hfinish.            
            
            assert (H16: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                              (ts + Z.of_nat nrange) \/
                    ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
        (ts + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
                                              (ts + Z.of_nat nrange) + 1).
            {
              assert (H17: monotonic_continuous (ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf))).
              {
                apply TAC_pool_all_monotonic_continuous; auto; try omega.
              }
              apply H17.
              omega.
            }

            destruct H16 as [Heq | Hinc]; auto.
            {
              rewrite Heq; auto.
              omega.
            }
            {
              rewrite Hinc.
              
              rewrite vertical_map_canonical; try omega.
              assert (H18: ZMap.get fromid (TAC_pool_at_t te' (Z.to_nat N0) conf)
        ((ts + Z.of_nat nrange) / Tf * Tf) = ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                                                      ((ts + Z.of_nat nrange) / Tf * Tf)).
              {              
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }

              rewrite H18.

              destruct (zlt
              (ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                 (ts + Z.of_nat nrange) -
               ZMap.get fromid (TAC_pool_at_t (ts + Z.of_nat nrange) (Z.to_nat N0) conf)
                 ((ts + Z.of_nat nrange) / Tf * Tf)) Cf); try discriminate.

              rewrite Z.min_l; try omega.
              rewrite Z.min_l; try omega.              
            }
          }
        }
      }
    }
  Qed.      
  
  Lemma count_occurrence_overlapping_le_aux_ts:
    forall Nnat N0 id ts te te' conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 ->
      0 <= ts <= te ->
      te <= te' ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      workload_overlapping_remain_total Nnat id ts te te' te conf (TAC_pool_at_t te' (Z.to_nat N0) conf) <= count_type_occurrence2_aux Nnat N0 N0 ts te conf id te Sched_overlap.
  Proof.
    intros * HNnat HN0 Hts Hte' Hconfvalid.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid').
      simpl.
      
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.
        rewrite count_single_task_self_sched_overlap_is_0; try omega.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.
        destruct (is_overlapping (Z.of_nat Nnat) Tf id ts te te') eqn:Hoverlap.
        {
          assert (H: forall a b c d, a <= c -> b <= d -> a + b <= c + d). intros. omega.
          apply H; auto.
          remember (Z.to_nat (te - ts)) as nrange.
          assert (HTT: te = ts+Z.of_nat nrange).
          { rewrite Heqnrange. rewrite Z2Nat.id; omega. }
          rewrite HTT in *. clear HTT Heqnrange.

          assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.          
          assert (H3: 1 < Tf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H4: 0 < Cf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }

          assert (H20: (ts + Z.of_nat nrange) / Tf * Tf < ts + Z.of_nat nrange <=
                  ts + Z.of_nat nrange).
          {
            pose proof (is_overlapping_true_end_not_aligned _ _ _ _ _ _ Hoverlap).
            assert ((ts + Z.of_nat nrange) / Tf * Tf <= ts + Z.of_nat nrange).
            { apply divide_mul_le; omega. }
            omega.
          }          
          
          apply is_overlapping_then_single_task_type_count_le_ts; auto; try omega.
        }
        {
          assert (0 <= count_single_task_type_occurrence N0 N0 (Z.of_nat Nnat) ts te conf id te Sched_overlap).
          {
            apply count_single_task_type_occurrence_nonneg.
          }
          omega.
        }
      }
    }
  Qed.
    
  Lemma count_occurrence_overlapping_le_ts:
    forall Nnat id ts te te' conf,
      0 < Z.of_nat Nnat ->
      0 <= ts <= te ->
      te <= te' ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      workload_overlapping_remain_total Nnat id ts te te' te conf (TAC_pool_at_t te' Nnat conf) <=
      count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) ts te conf id te Sched_overlap.
  Proof.
    intros * HN Hts Hte Hconfvalid.

    rewrite count_type_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    assert (Hw: workload_overlapping_remain_total Nnat id ts te te' te conf (TAC_pool_at_t te' Nnat conf) = workload_overlapping_remain_total Nnat id ts te te' te conf (TAC_pool_at_t te' (Z.to_nat (Z.of_nat Nnat)) conf)).
    {
      rewrite H1. auto.
    }

    rewrite Hw.
    apply count_occurrence_overlapping_le_aux_ts; auto; try omega.
  Qed.      
  
End SCHED_OVERLAP.



Section SCHEDULABLE_CONF.

  Theorem all_components_fit_if_schedulable_aux_v:
    forall Nnat conf id k T C np v,
      Z.of_nat np * k * T <= v <= Z.of_nat np * k * T + k * T ->
      system_schedulable (Z.of_nat Nnat) conf ->
      0 <= id < Z.of_nat Nnat ->      
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn : Z, 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= k * T) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      workload_overlapping_total Nnat id v (Z.of_nat np * (k * T) + k*T) (Z.of_nat np * (k * T) + k*T + k*T) conf -
      workload_remain_total Nnat id (Z.of_nat (S np) * (k * T)) 
                            (Z.of_nat (S np) * (k * T) + (k * T)) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * (k * T) + (k * T)) Nnat conf) + 
      workload_total1 Nnat id v
                      (Z.of_nat np * (k * T) + (k * T)) conf (TAC_pool_at_t (Z.of_nat np * (k * T) + (k * T)) Nnat conf) +
      workload_leftover_total Nnat id v (Z.of_nat np * (k * T) + k*T) conf <=
      Z.of_nat np * (k * T) + k * T - v - (Z.of_nat np * (k * T) + k * T - v) / T * C.
  Proof.
    intros * Hv Hschedulable Hid Hconfvalid Hlarge Hconfid.
    assert (HTC: 1 < T /\ 0 < C).
    {
      specialize (Hconfvalid _ _ _ Hid Hconfid).
      omega.
    }
    assert (HkT: T <= k * T).
    {
      specialize (Hlarge _ _ _ Hid Hconfid).
      omega.
    }

    assert (Hcount_total: count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id (Z.of_nat np * (k * T) + k * T) Sched_self +
                          count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id (Z.of_nat np * (k * T) + k * T) Sched_interference +
                          count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id (Z.of_nat np * (k * T) + k * T) Sched_leftover +
                          count_type_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id (Z.of_nat np * (k * T) + k * T) Sched_overlap <=                          Z.of_nat np * (k * T) + k * T - v).
    {
      apply sched_type_is_bounded_by_window; try nia.      
    }

    assert (Hmod1: (Z.of_nat np * (k * T) + k * T) mod T = 0).
    {
      replace (Z.of_nat np * (k * T) + k * T) with ((Z.of_nat np * k + k) * T) by lia.
      apply Z_mod_mult.
    }
    
    assert (Hself: (Z.of_nat np * (k * T) + k * T - v) / T * C <=
                   count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id
                   (Z.of_nat np * (k * T) + k * T) Sched_self).
    {
      apply count_occurrence_self_more_than_budget_multiple_periods; auto; try nia.
    }
    
    assert (Hinterference: count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id
                   (Z.of_nat np * (k * T) + k * T) Sched_interference =
           workload_total1 Nnat id v (Z.of_nat np * (k * T) + k * T) conf
     (TAC_pool_at_t (Z.of_nat np * (k * T) + k * T) Nnat conf)).
    {
      rewrite workload_total_eq.      
      apply count_occurrence_interference_equals; auto; try nia.
    }

    assert (Hleftover: count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id
                   (Z.of_nat np * (k * T) + k * T) Sched_leftover >=
                       workload_leftover_total Nnat id v (Z.of_nat np * (k * T) + k * T) conf).
    {
      apply count_occurrence_leftover_more; auto; try nia.
      - intros * Hi Hconfi.
        specialize (Hlarge _ _ _ Hi Hconfi).
        nia.
    }

    assert (Hoverlap: workload_overlapping_total Nnat id v (Z.of_nat np * (k * T) + k * T)
     (Z.of_nat np * (k * T) + k * T + k * T) conf -
   workload_remain_total Nnat id (Z.of_nat (S np) * (k * T))
     (Z.of_nat (S np) * (k * T) + k * T) conf
     (TAC_pool_at_t (Z.of_nat (S np) * (k * T) + k * T) Nnat conf) <=
                      count_type_occurrence (Z.of_nat Nnat) 
                   (Z.of_nat Nnat) v (Z.of_nat np * (k * T) + k * T) conf id
                   (Z.of_nat np * (k * T) + k * T) Sched_overlap).
    {
      assert (H: Z.of_nat (S np) = Z.of_nat np + 1). xomega.
      rewrite H.
      replace ((Z.of_nat np + 1) * (k * T)) with (Z.of_nat np * (k * T) + k * T) by lia.

      assert (H2: workload_remain_total Nnat id (Z.of_nat np * (k * T) + k * T) 
                                        (Z.of_nat np * (k * T) + k * T + k * T) conf (TAC_pool_at_t (Z.of_nat np * (k * T) + k * T + k * T) Nnat conf) =
             workload_remain_total Nnat id (Z.of_nat np * (k * T) + k * T)
                                   (Z.of_nat np * (k * T) + k * T + k * T) conf (TAC_pool_at_t (Z.of_nat np * (k * T) + k * T + k * T) (Z.to_nat (Z.of_nat Nnat)) conf)).
      {
        rewrite Nat2Z.id. auto.
      }
      rewrite H2.

      transitivity (workload_overlapping_remain_total Nnat id v (Z.of_nat np * (k * T) + k * T)
         (Z.of_nat np * (k * T) + k * T + k * T) (Z.of_nat np * (k * T) + k * T) conf
     (TAC_pool_at_t (Z.of_nat np * (k * T) + k * T + k * T) (Z.to_nat (Z.of_nat Nnat)) conf)).
      {
        apply workload_overlapping_remain_total_le_ts; auto; try nia.
      }
      {
        rewrite Nat2Z.id.
        apply count_occurrence_overlapping_le_ts; auto; try nia.
      }
    }
    omega.    
  Qed.    

  
  Corollary all_components_fit_if_schedulable:
    forall Nnat conf id k T C np,
      system_schedulable (Z.of_nat Nnat) conf ->
      0 <= id < Z.of_nat Nnat ->      
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn : Z, 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= k * T) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      workload_overlapping_total Nnat id (Z.of_nat np * (k * T)) (Z.of_nat np * (k * T) + k*T) (Z.of_nat np * (k * T) + k*T + k*T) conf -
      workload_remain_total Nnat id (Z.of_nat (S np) * (k * T)) 
                            (Z.of_nat (S np) * (k * T) + (k * T)) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * (k * T) + (k * T)) Nnat conf) + 
      workload_total1 Nnat id (Z.of_nat np * (k * T)) 
                      (Z.of_nat np * (k * T) + (k * T)) conf (TAC_pool_at_t (Z.of_nat np * (k * T) + (k * T)) Nnat conf) +
      workload_leftover_total Nnat id (Z.of_nat np * (k * T)) (Z.of_nat np * (k * T) + k*T) conf <=
      k * T - k * C.
  Proof.
    intros * Hschedulable Hid Hconf Hlarge Hconfid.
    assert (HT: 1 < T).
    {
      specialize (Hconf _ _ _ Hid Hconfid).
      omega.
    }
    assert (HkT: T <= k * T).
    {
      specialize (Hlarge _ _ _ Hid Hconfid).
      omega.
    }
    assert (H1: Z.of_nat np * k * T <= Z.of_nat np * (k * T) <= Z.of_nat np * k * T + k * T). nia.
    pose proof (all_components_fit_if_schedulable_aux_v _ _ _ _ _ _ _ _ H1 Hschedulable Hid Hconf Hlarge Hconfid) as Hbound.

    assert (H2:  Z.of_nat np * (k * T) + k * T - Z.of_nat np * (k * T) -
                                         (Z.of_nat np * (k * T) + k * T - Z.of_nat np * (k * T)) / T * C = k * T - k * C).
    {
      replace (Z.of_nat np * (k * T) + k * T - Z.of_nat np * (k * T)) with (k*T) by omega.
      rewrite Z_div_mult_full; try omega.
    }
    rewrite H2 in Hbound.
    auto.
  Qed.    

  Corollary all_components_fit_if_schedulable_v:
    forall Nnat conf id k T C np v,
      Z.of_nat np * k * T < v <= Z.of_nat np * k * T + k * T ->
      system_schedulable (Z.of_nat Nnat) conf ->
      0 <= id < Z.of_nat Nnat ->      
      config_pool_valid (Z.of_nat Nnat) conf ->
      (forall i Tn Cn : Z, 0 <= i < Z.of_nat Nnat -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= k * T) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      workload_overlapping_total Nnat id v (Z.of_nat np * (k * T) + k*T) (Z.of_nat np * (k * T) + k*T + k*T) conf -
      workload_remain_total Nnat id (Z.of_nat (S np) * (k * T)) 
                            (Z.of_nat (S np) * (k * T) + (k * T)) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * (k * T) + (k * T)) Nnat conf) + 
      workload_total1 Nnat id v
                      (Z.of_nat np * (k * T) + (k * T)) conf (TAC_pool_at_t (Z.of_nat np * (k * T) + (k * T)) Nnat conf) +
      workload_leftover_total Nnat id v (Z.of_nat np * (k * T) + k*T) conf <=
      Z.of_nat np * (k * T) + k * T - v.
  Proof.
    intros * Hv Hschedulable Hid Hconf Hlarge Hconfid.
    assert (H1: Z.of_nat np * k * T <= v <= Z.of_nat np * k * T + k * T). nia.
    assert (HTC: 1 < T /\ 0 < C).
    {
      specialize (Hconf _ _ _ Hid Hconfid).
      omega.
    }
    assert (HkT: T <= k * T).
    {
      specialize (Hlarge _ _ _ Hid Hconfid).
      omega.
    }
    
    pose proof (all_components_fit_if_schedulable_aux_v _ _ _ _ _ _ _ _ H1 Hschedulable Hid Hconf Hlarge Hconfid) as Hbound.
    transitivity (Z.of_nat np * (k * T) + k * T - v -
           (Z.of_nat np * (k * T) + k * T - v) / T * C); auto.
    {
      assert (H2: 0 <= (Z.of_nat np * (k * T) + k * T - v) / T * C).
      {
        apply div_mul_nonneg; try nia.
      }
      omega.
    }
  Qed.

  Corollary workloads_bounded_if_schedulable_self:
    forall N conf id k T C p,
      0 <= p ->
      system_schedulable N conf ->
      0 <= id < N ->      
      config_pool_valid N conf ->
      (forall i Tn Cn : Z, 0 <= i < N -> ZMap.get i conf = mkPrioConfigValid Tn Cn -> Tn <= k * T) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      workload_total1 (Z.to_nat N) id (p * (k * T))
                      (p * (k * T) + k * T) conf (TAC_pool_at_t (p * (k * T) + k * T) (Z.to_nat N) conf) <=
      k * T - k * C.
  Proof.
    intros * Hp Hsched Hid Hconfvalid Hlarge Hconfid.
    assert (HT: 1 < T /\ 0 < C).
    {
      specialize (Hconfvalid _ _ _ Hid Hconfid).
      omega.
    }
    assert (HkT: T <= k * T).
    {
      specialize (Hlarge _ _ _ Hid Hconfid).
      omega.
    }
    assert (H1: p * k * T <= p * (k * T) <= p * k * T + k * T). nia.

    assert (Hcount_total: count_type_occurrence N N (p * (k * T)) (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_self +
                          count_type_occurrence N N (p * (k * T)) (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_interference +
                          count_type_occurrence N N (p * (k * T)) (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_leftover +
                          count_type_occurrence N N (p * (k * T)) (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_overlap <= (p * (k * T) + k * T) - p * (k * T)).
    {
      apply sched_type_is_bounded_by_window; try nia.                                                  }

    assert (Hmod: (p * (k * T) + k * T) mod T = 0).
    {
      replace (p * (k * T) + k * T) with ((p * k + k) * T) by lia.
      apply Z_mod_mult.
    }
    
    assert (Hself: (p * (k * T) + k * T - p * (k * T)) / T * C <=
                    count_type_occurrence N N (p * (k * T)) (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_self).
    {
      apply count_occurrence_self_more_than_budget_multiple_periods; auto; try nia.
    }
    replace (p * (k * T) + k * T - p * (k * T)) with (k * T) in Hself by omega.
    rewrite Z_div_mult_full in Hself; try omega.

    assert (Hinterference: count_type_occurrence N N (p * (k * T)) 
                                                 (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_interference =
           workload_total1 (Z.to_nat N) id (p * (k * T)) (p * (k * T) + k * T) conf
     (TAC_pool_at_t (p * (k * T) + k * T) (Z.to_nat N) conf)).
    {
      rewrite workload_total_eq.

      remember (Z.to_nat N) as Nnat.
      assert (HNN: N = Z.of_nat Nnat).
      {
        rewrite HeqNnat.
        rewrite Z2Nat.id; omega.
      }
      rewrite HNN in *.
      
      apply count_occurrence_interference_equals; auto; try nia.
    }

    assert (H2: 0 <= count_type_occurrence N N (p * (k * T)) 
                                          (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_leftover).
    {
      apply count_type_occurrence_nonneg.
    }

    assert (H3: 0 <= count_type_occurrence N N (p * (k * T)) 
                                          (p * (k * T) + k * T) conf id (p * (k * T) + k * T) Sched_overlap).
    {
      apply count_type_occurrence_nonneg.
    }

    omega.
  Qed.


  Corollary workloads_bounded_if_schedulable_others:
    forall N conf i T C p,
      0 <= p ->
      system_schedulable N conf ->
      0 <= i < N ->      
      config_pool_valid N conf ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      workload_total1 (Z.to_nat N) i (p * T)
                      (p * T + T) conf (TAC_pool_at_t (p * T + T) (Z.to_nat N) conf) <=
      T - C.
  Proof.
    intros * Hp Hsched Hi Hconfvalid Hconfi.
    assert (HT: 1 < T /\ 0 < C).
    {
      specialize (Hconfvalid _ _ _ Hi Hconfi).
      omega.
    }

    assert (Hcount_total: count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_self +
                          count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_interference +
                          count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_leftover +
                          count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_overlap <= p * T + T - p * T).
    {
      apply sched_type_is_bounded_by_window; try nia.                                                  }

    assert (Hmod: (p * T + T) mod T = 0).
    {
      replace (p * T + T) with ((p + 1) * T) by lia.
      apply Z_mod_mult.
    }
    
    assert (Hself: (p * T + T - p * T) / T * C <=
                    count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_self).
    {
      apply count_occurrence_self_more_than_budget_multiple_periods; auto; try nia.
    }
    replace (p * T + T - p * T) with T in Hself by omega.
    rewrite Z_div_same_full in Hself; try omega.    

    assert (Hinterference: count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_interference =
          workload_total1 (Z.to_nat N) i (p * T) (p * T + T) conf
     (TAC_pool_at_t (p * T + T) (Z.to_nat N) conf)).
    {
      rewrite workload_total_eq.

      remember (Z.to_nat N) as Nnat.
      assert (HNN: N = Z.of_nat Nnat).
      {
        rewrite HeqNnat.
        rewrite Z2Nat.id; omega.
      }
      rewrite HNN in *.
      
      apply count_occurrence_interference_equals; auto; try nia.
    }

    assert (H2: 0 <= count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_leftover).
    {
      apply count_type_occurrence_nonneg.
    }

    assert (H3: 0 <= count_type_occurrence N N (p * T) (p * T + T) conf i (p * T + T) Sched_overlap).
    {
      apply count_type_occurrence_nonneg.
    }

    omega.
  Qed.
  
  
End SCHEDULABLE_CONF.


