Require Import Coqlib.
Require Import Maps.
Require Import VTDefs.
Require Import Integers.
Require Import Values.
Require Import LemmaVTMap.
Require Import LemmaInterference.
Require Import LemmaOtherWorkload.
Require Import LemmaOtherWorkload1.
Require Import LemmaSchedulableWorkload.
Require Import LemmaMaths.
Require Import Coq.Sorting.Mergesort.
Require Import Coq.Structures.Orders.
Require Import Coq.Sorting.Permutation.
Require Import Psatz.


Local Open Scope Z_scope.

Section ENLARGE.

  Definition enlarged_task_has_largest_period id k N conf :=
    forall i T C Ti Ci,
      0 <= i < N ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      ZMap.get i conf = mkPrioConfigValid Ti Ci ->
      Ti <= k * T.

  
End ENLARGE.

Section WORKLOADWHOLE.
  
  Lemma workload_whole_total_on_self_remains_same_after_enlarge:
    forall p conf confe Nnat id k T C,
      0 <= p ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      id_is_enlarged_by_k id k conf confe ->
      workload_whole_total Nnat id (p * k * T) (p * k * T + k * T) confe =
      workload_whole_total Nnat id (p * k * T) (p * k * T + k * T) conf.
  Proof.
    intros * Hp Hconfid Henlarge.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHNnat.
      destruct (zeq id (Z.of_nat Nnat)); auto. 
      assert (Hconfeq: ZMap.get (Z.of_nat Nnat) confe = ZMap.get (Z.of_nat Nnat) conf).
      {
        unfold id_is_enlarged_by_k in Henlarge.
        erewrite Henlarge; eauto.
        rewrite ZMap.gso; auto.
      }
      rewrite Hconfeq.
      auto.
    }
  Qed.

  Lemma workload_whole_total_on_self_remains_same_after_enlarge_generalized:
    forall conf confe Nnat id k T C ts te,
      ZMap.get id conf = mkPrioConfigValid T C ->
      id_is_enlarged_by_k id k conf confe ->
      workload_whole_total Nnat id ts te confe =
      workload_whole_total Nnat id ts te conf.
  Proof.
    intros * Hp Hconfid Henlarge.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHNnat.
      destruct (zeq id (Z.of_nat Nnat)); auto. 
      assert (Hconfeq: ZMap.get (Z.of_nat Nnat) confe = ZMap.get (Z.of_nat Nnat) conf).
      {
        unfold id_is_enlarged_by_k in Henlarge.
        erewrite Henlarge; eauto.
        rewrite ZMap.gso; auto.
      }
      rewrite Hconfeq.
      auto.
    }
  Qed.
  
End WORKLOADWHOLE.


Section WORKLOAD_REMAIN_BUSY.

  Lemma workload_remain_total_non_inc_next_period_step_when_busy:
    forall N conf conf2 id k T C np,
      1 < k ->
      1 < T ->
      0 < C ->
      0 <= id < N ->
      config_pool_valid N conf ->
      config_pool_valid N conf2 ->
      id_is_enlarged_by_k id k conf conf2 ->
      enlarged_task_has_largest_period id k N conf ->
      system_schedulable N conf ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf2
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf2) <=
      workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf) ->
      last_idle_slot N N (Z.of_nat np * k * T) 
                     (Z.of_nat np * k * T + k * T) conf2 = None ->
      workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                            (Z.of_nat (S np) * k * T + k * T) conf2
                            (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf2) <=
      workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                            (Z.of_nat (S np) * k * T + k * T) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf).
  Proof.
    intros * Hk HT HC Hid Hconf Hconf2 Henlarged Hperiod Hschedulable Hconfid Hprevvalid Hidle.
    assert (HNN: Z.of_nat (Z.to_nat N) = N).
    {
      rewrite Z2Nat.id; omega.
    }
    
    assert (HkT: 1 < k * T). nia.
    assert (HkC: 0 < k * C). nia.
    assert (H1: Z.of_nat np * k * T = Z.of_nat np * (k * T)). lia.
    assert (H2: Z.of_nat (S np) * k * T = Z.of_nat (S np) * (k * T)). lia.
    assert (Hid': 0 <= id < Z.of_nat (Z.to_nat N)). omega.
    assert (Hconf2id: ZMap.get id conf2 = mkPrioConfigValid (k * T) (k * C)).
    {
      specialize (Henlarged id _ _ Hconfid).
      rewrite ZMap.gss in Henlarged.
      auto.
    }

    assert (Hperiod2: forall i Tn Cn : Z, 0 <= i < N -> ZMap.get i conf2 = mkPrioConfigValid Tn Cn -> Tn <= k * T).
    {
      intros * Hi Hconfi.
      rewrite (Henlarged _ _ _ Hconfid) in Hconfi.
      destruct (zeq i id) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite ZMap.gss in Hconfi.
        inversion Hconfi.
        omega.
      }
      {
        rewrite ZMap.gso in Hconfi; auto.
        specialize (Hperiod _ _ _ _ _ Hi Hconfid Hconfi).
        auto.
      }
    }        
    
    assert (Hbound2: workload_overlapping_total (Z.to_nat N) id (Z.of_nat np * k * T) 
                       (Z.of_nat np * k * T + k*T) (Z.of_nat np * k * T + k*T + k*T) conf2 -
                     workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                                           (Z.of_nat (S np) * k * T + k * T) conf2
                                           (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf2) >= k * T - k * C -  workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf2
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf2) -
                                                                                                                  workload_leftover_total (Z.to_nat N) id (Z.of_nat np * k * T) (Z.of_nat np * k * T + k*T) conf2).
    {
      rewrite H1, H2.
      apply interference_is_bounded_when_schedule_is_busy_single_period; auto.
      - rewrite HNN.
        auto.
      - rewrite HNN.
        auto.
      - rewrite HNN.
        replace (Z.of_nat np * (k * T)) with (Z.of_nat np * k * T) by lia.
        auto.
    }
      
    assert (Hleftover:  workload_leftover_total (Z.to_nat N) id (Z.of_nat np * k * T) (Z.of_nat np * k * T + k*T) conf2 =  workload_leftover_total (Z.to_nat N) id (Z.of_nat np * k * T) (Z.of_nat np * k * T + k*T) conf).
    { erewrite workload_leftover_total_same; eauto. }
    assert (Hoverlap: workload_overlapping_total (Z.to_nat N) id (Z.of_nat np * k * T) (Z.of_nat np * k * T + k*T) (Z.of_nat np * k * T + k*T + k*T) conf2 = workload_overlapping_total (Z.to_nat N) id (Z.of_nat np * k * T) (Z.of_nat np * k * T + k*T) (Z.of_nat np * k * T + k*T + k*T) conf).
    { erewrite workload_overlapping_total_same; eauto. }
    assert (Hbound1: workload_overlapping_total (Z.to_nat N) id (Z.of_nat np * k * T) 
                       (Z.of_nat np * k * T + k*T) (Z.of_nat np * k * T + k*T + k*T) conf -
                     workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                                           (Z.of_nat (S np) * k * T + k * T) conf
                                           (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf) + workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf) +
                     workload_leftover_total (Z.to_nat N) id (Z.of_nat np * k * T) (Z.of_nat np * k * T + k*T) conf <=
                     k * T - k * C).
    {
      rewrite H1, H2.
      apply all_components_fit_if_schedulable; auto.
      - rewrite HNN. auto.
      - rewrite HNN. auto.
      - rewrite HNN.
        unfold enlarged_task_has_largest_period in Hperiod.
        intros * Hi Hconfi.
        specialize (Hperiod _ _ _ _ _ Hi Hconfid Hconfi).
        auto.
    }
    omega.
  Qed.
  
End WORKLOAD_REMAIN_BUSY.


Section WORKLOAD_REMAIN_NOT_BUSY.

  Lemma workload_remain_total_non_inc_next_period_step_when_not_busy:
    forall N conf conf2 id k T C np v,
      1 < k ->
      0 <= id < N ->
      config_pool_valid N conf ->
      config_pool_valid N conf2 ->
      id_is_enlarged_by_k id k conf conf2 ->
      enlarged_task_has_largest_period id k N conf ->
      system_schedulable N conf ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf2
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf2) <=
      workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf) ->
      last_idle_slot N N (Z.of_nat np * k * T) 
                     (Z.of_nat np * k * T + k * T) conf2 = Some v ->
      workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                            (Z.of_nat (S np) * k * T + k * T) conf2
                            (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf2) <=
      workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                            (Z.of_nat (S np) * k * T + k * T) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf).
  Proof.
    intros * Hk Hid Hconf Hconf2 Henlarge Hperiod Hschedulable Hconfid Hhypothesis Hidle.

    assert (HT: 1 < T).
    {
      specialize (Hconf _ _ _ Hid Hconfid).
      omega.
    }
    assert (HC: 0 < C).
    {
      specialize (Hconf _ _ _ Hid Hconfid).
      omega.
    }
    
    assert (H0: 1 <= k * T). nia.
    assert (H1: Z.of_nat np * k * T <= Z.of_nat np * k * T + k * T). omega.
    pose proof (last_idle_slot_range _ _ _ _ _ _ H1 Hidle) as Hidlerange.
    assert (H3: Z.of_nat np * k * T = Z.of_nat np * (k * T)). lia.
    assert (H4: Z.of_nat (S np) * k * T = Z.of_nat (S np) * (k * T)). lia.
    assert (HNN: Z.of_nat (Z.to_nat N) = N).
    {
      rewrite Z2Nat.id; omega.
    }

    assert (Hperiod2: forall i Tn Cn : Z, 0 <= i < N -> ZMap.get i conf2 = mkPrioConfigValid Tn Cn -> Tn <= k * T).
    {
      intros * Hi Hconfi.
      rewrite (Henlarge _ _ _ Hconfid) in Hconfi.
      destruct (zeq i id) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite ZMap.gss in Hconfi.
        inversion Hconfi.
        omega.
      }
      {
        rewrite ZMap.gso in Hconfi; auto.
        specialize (Hperiod _ _ _ _ _ Hi Hconfid Hconfi).
        auto.
      }
    }            

    assert (Hr2: workload_overlapping_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k * T) (Z.of_nat np * k * T + k * T + k * T) conf2 - workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
     (Z.of_nat (S np) * k * T + k * T) conf2
     (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf2)                 
                 >= Z.of_nat np * k * T + k * T - (v+1) - workload_leftover_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf2
                 - workload_whole_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf2
           ).
    {
      rewrite H3, H4.
      assert (Hconf2id: ZMap.get id conf2 = mkPrioConfigValid (k*T) (k*C)).
      {
        unfold id_is_enlarged_by_k in Henlarge.
        rewrite Henlarge with (T:=T) (C:=C); auto.
        rewrite ZMap.gss.
        auto.
      }
      apply interference_is_bounded_when_schedule_is_busy_last_idle with (C:=k*C); auto; try nia.
      - rewrite HNN. auto.
      - rewrite HNN. auto.
      - rewrite HNN. replace (Z.of_nat np * (k * T)) with (Z.of_nat np * k * T) by lia.
        auto.
    }

    assert (Hleftover: workload_leftover_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf2 = workload_leftover_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf).
    { erewrite workload_leftover_total_same; eauto. }

    assert (Hoverlapping: workload_overlapping_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k * T) (Z.of_nat np * k * T + k * T + k * T) conf2 = workload_overlapping_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k * T) (Z.of_nat np * k * T + k * T + k * T) conf).
    { erewrite workload_overlapping_total_same; eauto. }

    assert (Hwhole: workload_whole_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf2 = workload_whole_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf).
    {
      eapply workload_whole_total_on_self_remains_same_after_enlarge_generalized; eauto.
    }
    
    assert (Hbound1: workload_overlapping_total (Z.to_nat N) id (v+1)
                       (Z.of_nat np * k * T + k*T) (Z.of_nat np * k * T + k*T + k*T) conf -
                     workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                                           (Z.of_nat (S np) * k * T + k * T) conf
                                           (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf) + workload_total1 (Z.to_nat N) id (v+1)
                      (Z.of_nat np * k * T + k * T) conf
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf) +
                     workload_leftover_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf <=
                     Z.of_nat np * k * T + k * T - (v+1)).
    {
      rewrite H3, H4.
      apply all_components_fit_if_schedulable_v with (C:=C); auto; try omega.
      - rewrite HNN. auto.
      - rewrite HNN. auto.
      - rewrite HNN.
        unfold enlarged_task_has_largest_period in Hperiod.
        intros * Hi Hconfi.
        specialize (Hperiod _ _ _ _ _ Hi Hconfid Hconfi).
        auto.
    }

    unfold workload_total1 in Hbound1.
    assert (0 <= workload_remain_total (Z.to_nat N) id 
               (v + 1) (Z.of_nat np * k * T + k * T) conf
               (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf)).
    {
      apply workload_remain_total_nonneg.
    }
    assert (Hr3: workload_overlapping_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k * T) (Z.of_nat np * k * T + k * T + k * T) conf - workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
     (Z.of_nat (S np) * k * T + k * T) conf
     (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf)                 
                 <= Z.of_nat np * k * T + k * T - (v+1) - workload_leftover_total (Z.to_nat N) id (v+1) (Z.of_nat np * k * T + k*T) conf - workload_whole_total (Z.to_nat N) id (v + 1) (Z.of_nat np * k * T + k * T) conf). omega.
    omega.
  Qed.

End WORKLOAD_REMAIN_NOT_BUSY.


Section WORKLOADREMAIN.

  Lemma workload_remain_total_non_inc_next_period_step:
    forall N conf conf2 id k T C np,
      1 < k ->
      1 < T ->
      0 < C ->
      0 <= id < N ->
      config_pool_valid N conf2 ->
      config_pool_valid N conf ->
      id_is_enlarged_by_k id k conf conf2 ->
      enlarged_task_has_largest_period id k N conf ->
      system_schedulable N conf ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf2
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf2) <=
      workload_total1 (Z.to_nat N) id (Z.of_nat np * k * T) 
                      (Z.of_nat np * k * T + k * T) conf
                      (TAC_pool_at_t (Z.of_nat np * k * T + k * T) (Z.to_nat N) conf) ->
      workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                            (Z.of_nat (S np) * k * T + k * T) conf2
                            (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf2) <=
      workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
                            (Z.of_nat (S np) * k * T + k * T) conf
                            (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf).
  Proof.
    intros * Hk HT HC Hid Hconf2 Hconf Henlarged Hperiod Hschedulable Hconfid Hprevvalid.
    destruct (last_idle_slot N N (Z.of_nat np * k * T) (Z.of_nat np * k * T + k * T) conf2) as [Hidle |] eqn:Hlast.
    {
      apply workload_remain_total_non_inc_next_period_step_when_not_busy with (C:=C) (v:=Hidle); auto.
    }
    {
      apply workload_remain_total_non_inc_next_period_step_when_busy with (C := C); auto; try omega.
    }
  Qed.
  
End WORKLOADREMAIN.


Section MAIN.

  Theorem enlarge_does_not_incur_more_workload_on_self:
    forall N conf conf2 id k T C,
      1 < k ->
      0 <= id < N ->
      config_pool_valid N conf ->
      config_pool_valid N conf2 ->
      id_is_enlarged_by_k id k conf conf2 ->
      enlarged_task_has_largest_period id k N conf ->
      system_schedulable N conf ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      (forall p, 0 <= p ->
          workload_total1 (Z.to_nat N) id (p * k * T) (p * k * T + k * T) conf2
                            (TAC_pool_at_t (p * k * T + k * T) (Z.to_nat N) conf2) <=
            workload_total1 (Z.to_nat N) id (p * k * T) (p * k * T + k * T) conf
                            (TAC_pool_at_t (p * k * T + k * T) (Z.to_nat N) conf)).
  Proof.
    intros * Hk Hid Hconfvalid Hconf2valid Henlarged Hperiod Hschedulable Hconfid p Hp.
    remember (Z.to_nat p) as np.
    assert (Hpp: p = Z.of_nat np).
    { rewrite Heqnp. rewrite Z2Nat.id; omega. }
    clear Heqnp.
    rewrite Hpp in *. clear Hpp.
    assert (HT: 1 < T).
    { specialize (Hconfvalid _ _ _ Hid Hconfid). omega. }
    induction np.
    {
      (* base case, the first period *)
      unfold workload_total1.
      replace (Z.of_nat 0) with 0 by (simpl; auto).
      rewrite workload_whole_total_on_self_remains_same_after_enlarge with (conf := conf) (k := k) (C := C); auto.
      simpl.
      repeat rewrite workload_remain_total_0_at_0.
      omega.
    }
    {
      (* induction *)
      assert (HC: 0 < C).
      { specialize (Hconfvalid _ _ _ Hid Hconfid). omega. }      
      unfold workload_total1.
      rewrite workload_whole_total_on_self_remains_same_after_enlarge with (conf := conf) (k := k) (C := C); auto.
      assert (Hremain: workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
     (Z.of_nat (S np) * k * T + k * T) conf2
     (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf2) <=
             workload_remain_total (Z.to_nat N) id (Z.of_nat (S np) * k * T)
     (Z.of_nat (S np) * k * T + k * T) conf
     (TAC_pool_at_t (Z.of_nat (S np) * k * T + k * T) (Z.to_nat N) conf)).
      {
        apply workload_remain_total_non_inc_next_period_step with (C := C); auto.
        apply IHnp.
        omega.
      }
      omega.
    }
  Qed.
      
End MAIN.

